"Generic Interpreter Module"

# ***************** ApplicationBuffer **************************

class ApplicationBuffer:
    "Its objects represent the application code buffer to be recognised"

    def __init__(self,code):
        # Application code
        self.code=code
        # Character position being recognised in the buffer 
        self.position=0
        # Text post-recognition
        self.text=""

# ***************** Generic Interpreter **************************

class GI:
    "Generic Interpreter that evaluates and application in the nitrO system"
    
    def __init__(self,application,nitrO):
        "Generic Interpreter Constructor"
        # Application to be evaluated
        self.application=application
        # Application Buffer
        self.buffer=ApplicationBuffer(application.sourceCode)
        # Global nitrO Facade object
        self.nitrO=nitrO

    def objectsCreation(self):
        "Method that creates the object grammar and the object application tree. Returns the GenericInterpreter instance (text, application...)"
        startRule=self.application.language.start
        rootNode=startRule.recognise(self)
        if not(rootNode):
            raise "Application not validated."
        else:
            self.application.rootNode=rootNode
        self.clearBuffer()
        if self.buffer.position!=len(self.buffer.code):
            raise "Application not validated."
        return self
    
    def evaluate(self):
        "Method that executes the application."
        self.objectsCreation()
        self.application.isRunning=1
        self.application.window.setTitle(self.application.describe())
        self.application.rootNode.execute()
        self.application.isRunning=0
        self.application.window.setTitle(self.application.describe())
        return self

    def clearSkip(self):
        "Clear every 'Skip' token in the source code at the current position"
        tokensRecognised=1
        while tokensRecognised:
            tokensRecognised=0
            for token in self.application.language.skipSpec:
                initialText=self.buffer.text
                if token.recognise(self):
                    # Skip it!
                    self.buffer.text=initialText
                    tokensRecognised=tokensRecognised+1

    def clearNotSkip(self):
        "Takes every sequence of 'NotSkip' token and returns it"
        initialText=self.buffer.text
        tokensRecognised=1
        while tokensRecognised:
            tokensRecognised=0
            for token in self.application.language.notSkipSpec:
                if token.recognise(self):
                    tokensRecognised=tokensRecognised+1
        return self.buffer.text[len(initialText):]

    def clearBuffer(self):
        "Clears the 'Skip' tokens and takes the 'NotSkip' ones. Returns the sequence of 'NotSkip' tokens."
        text=""
        somethingRecognised=1
        while somethingRecognised:
            self.clearSkip()
            newText=self.clearNotSkip()
            somethingRecognised=newText!=""
            text=text+newText
            newText=""
        return text

# ******************** Module Testing Functions ********************

# Skip and NotSkip tokens Testing Function
import langSpec
import metaLang
global nitrO
def testSkip():
    "Skip and NotSkip tokens Testing Function"
    code=""" Application = "My sample application"
          Language = "Expression"
\t \t
                            print "Python Code" # Python Comment
            """
    nitrO.execute(code)
    gi=GI(nitrO.apps["My sample application"])
    s=gi.clearBuffer()
    print s
    return s
    
# Application evaluation testing function
def testEval():
    "Application evaluation testing function"
    code=""" Application = "My sample Expression application"
          Language = "Expression"
    34 +   89;
    2+ 34 - 76767 * 32 / 34232; 
            """
    gi=nitrO.execute(code)
    return gi.application.rootNode.text
    
# Application treeNodes creation testing function
def testTreeNodes():
    "Application treeNodes creation testing function"
    code=""" Application = "My sample Expression application"
          Language = "Expression"
    4;
"""
    gi=nitrO.execute(code)
    print gi.application.rootNode.describeTree()
    return gi
