Language    = MetaPython

Scanner = {
	"Digit Token"
		digit -> "0" | "1" | "2" | "3" | "4" | "5" | "6" | "7" | "8" | "9"
			;
	"Integer Token"
		INTEGER -> digit moreDigits
			;
	"Float Token"
		FLOAT -> INTEGER "." INTEGER
			;
	"Zero or more digits token"
		moreDigits -> digit moreDigits
			|
			;
	"Characacter Token"
		char -> "a" | "b" | "c" | "d" | "e" | "f" | "g" | "h" | "i" | "j" | "k" | "l" | "m" | "n" | "" |
			  "o" | "p" | "q" | "r" | "s" | "t" | "u" | "v" | "w" | "x" | "y" | "z" |
			  "" | "" | "" | "" | "" |
			  "A" | "B" | "C" | "D" | "E" | "F" | "G" | "H" | "I" | "J" | "K" | "L" | "M" | "N" | "" |
			  "O" | "P" | "Q" | "R" | "S" | "T" | "U" | "V" | "W" | "X" | "Y" | "Z" | "" | "" | "" | "" | "" |
			  "_" 
			;
	"Character or Digit Token"
		charOrDigit -> char
				| digit
				| "_"
				;
	"Row Char Token"
		rowChar -> char
			| digit
			| "_"  
			| " " | "," | "." | ":" | ";" | "" | "!" | "" | "?" | "#" | "*" | "/" | "$" | "%" | "&" 
			| "|" | "(" | ")" | "@" | "=" | "[" | "]" | "{" | "}" | "+" | "-" | ">" | "<"
			| "\t" | "\n" | "\\"
			;
	"ID Token"
		ID -> char moreCharsOrDigits
			;
	"Label Token"
		LABEL -> "#" char moreCharsOrDigits "#"
			;
	"String Token"
		STRING -> "'" rowChar moreRowChars "'" 
		    ;
	"Zero or more chars or digits token"
		moreCharsOrDigits -> charOrDigit moreCharsOrDigits
			|
			;
	"Zero or more row chars"
		moreRowChars -> rowChar moreRowChars  
			|
			;
	"COLON Token"
		COLON -> ":"
			;
	"DOT Token"
		DOT -> "."
			;
	"SEMICOLON Token"
		SEMICOLON -> ";"
			;
	"COMMA Token"
		COMMA -> ","
			;
	"OPENCURLYBRACE Token"
		OPENCURLYBRACE -> "{"
			;
	"CLOSECURLYBRACE Token"
		CLOSECURLYBRACE -> "}"
			;
	"OPENBRACKET Token"
		OPENBRACKET -> "["
			;
	"CLOSEBRACKET Token"
		CLOSEBRACKET -> "]"
			;
	"OPENBRACE Token"
		OPENBRACE -> "("
			;
	"CLOSEBRACE Token"
		CLOSEBRACE -> ")"
			;
	"PLUS Operator token"
		PLUS -> "+"
			;
	"MINUS Operator token"
		MINUS -> "-"
			;
	"TIMES Operator token"
		TIMES -> "*"
			;
	"BY Operator token"
		BY -> "/"
			;
	"LOWER Operator token"
		LOWER -> "<"
			;
	"LOWEREQUAL Operator token"
		LOWEREQUAL -> "<="
			;
	"GREATER Operator token"
		GREATER -> ">"
			;
	"GREATEREQUAL Operator token"
		GREATEREQUAL -> ">="
			;
	"EQUAL Operator token"
		EQUAL -> "=="
			;
	"NOTEQUAL Operator token"
		NOTEQUAL -> "!="
			;
	"ASIGN token"
		ASIGN -> "="
			;
	"CLASS keyword"
		CLASS -> "class"
			;
	"DEF keyword"
		DEF -> "def"
			;
	"IF keyword"
		IF -> "if"
			;
	"ELSE keyword"
		ELSE -> "else"
			;
	"WHILE keyword"
		WHILE -> "while"
			;
	"RETURN keyword"
		RETURN -> "return"
			;
	"PRINT keyword"
		PRINT -> "print"
			;
	"FROM keyword"
		FROM -> "from"
			;
	"IMPORT keyword"
		IMPORT -> "import"
			;
}


Parser = {
	"Initial Free-Context Rule"
		S -> statement moreStatements <#
# Initial Code Needed on the Application Execution
# Used in the application execution
global classes,functions,classAnalysed,functionAnalised,functionResult
global expressionToString,argsToText,getObjectName,asignArgsToParams,invokeFunction

classes={} # Classes Symbol Table
functions={} # Function Symbol Table
classAnalysed="" # To separate method and function definition
functionAnalysed=0 # To separate the function call and function definition
functionResult=None # Has the value to be returned in a function call

def expressionToString(value):
	"Converts the value to a string"
	return str(value)

def argsToText(args):
	"Converts a list of values into a comma separated value string"
	result=""
	for arg in args:
		result=result+","+expressionToString(arg)
	return result[1:]

def getObjectName(value):
	"Gets the left object in a expression"
	value=value+"."
	return value[:string.index(value,".")]

def asignArgsToParams(args,params):
	"Defines the funtion param list with its initial value list (args values)"
	# _arg Takes the args values
	for i in range(len(params)):
		_arg=args[i]
		exec("global "+params[i]+"\n"+params[i]+"=_arg\n")

def invokeFunction(functionName,args):
	"Invokes a builtin function or method"
	argsString=""
	for i in range(len(args)):
		argsString=argsString+",args["+str(i)+"]"
	if len(argsString)>0:
		argsString=argsString[1:]
	# Is an exec call?
	if functionName=="exec":
		exec(functionName+"("+argsString+")\n")
		return None
	else:
		exec("global _temp,"+getObjectName(functionName)+"\n_temp="+functionName+"("+argsString+")\n")
		return _temp


nodes[1].execute()
nodes[2].execute()
#>
			;

	"Zero or more statements"
		moreStatements -> statement moreStatements <#
nodes[1].execute()
nodes[2].execute()
#>
			|
			;

	"Python Statement" 
		statement -> class <#
nodes[1].execute() # Inserts the class into the ST
#>
			| function <#
# Indicates that the function is being analysed (not executed)
global functionAnalysed
functionAnalysed=1
nodes[1].execute() # Inserts the class into the ST
functionAnalysed=0
#>
			| print SEMICOLON <#
nodes[1].execute() # Statatement execution
#>
			| if <#
nodes[1].execute() # Statatement execution
#>
			| while <#
nodes[1].execute() # Statatement execution
#>
			| import SEMICOLON <#
nodes[1].execute() # Statatement execution
#>
			| _REIFY_ <#
#write("Trace> Reify code executed.\n") # NoComment=TraceExecution
nodes[1].execute()
#>
			| asignment SEMICOLON <#
nodes[1].execute() # Statatement execution
#>
			| functionCall SEMICOLON <#
nodes[1].execute() 
#>
			;

	"Class Specification"
		class -> CLASS ID OPENCURLYBRACE classBody CLOSECURLYBRACE <#
# Indicates which class is being analysed
global classAnalysed
classAnalysed=nodes[2].text
# Inserts into the SY the class name
exec("global "+nodes[2].text+"\nclass "+nodes[2].text+":\n\tpass\n\n_temp="+nodes[2].text+"\n")
classes[classAnalysed]=_temp
#write("Trace> Class "+classAnalysed+" inserted into the ST.\n") # NoComment=TraceExecution
nodes[4].execute()
classAnalysed=""
#>
			;

	"Class Body"
		classBody -> function moreFunctions <#
# Indicates that the function is being analysed (not executed)
global functionAnalysed
functionAnalysed=1
nodes[1].execute()
nodes[2].execute()
functionAnalysed=0
#>
			;

	"Zero or more functions"
		moreFunctions -> function moreFunctions <#
nodes[1].execute()
nodes[2].execute()
#>
			|
			;

	"Function Definition"
		function -> DEF ID params OPENCURLYBRACE functionBody CLOSECURLYBRACE <#
# Analysed or Executed?
if functionAnalysed:
	# Function definition (analysed)
	nodes[3].execute() # nodes[3].value saves the param list
	# Analysed, but function or method?
	if classAnalysed: # insert into the ST
		functions[classAnalysed+"."+nodes[2].text]=nodes[0]
		#write("Trace> Function "+classAnalysed+"."+nodes[2].text+" inserted into the ST.\n") # NoComment=TraceExecution
	else:
		functions[nodes[2].text]=nodes[0]
		#write("Trace> Function "+nodes[2].text+" inserted into the ST.\n") # NoComment=TraceExecution
else:
	# Function Call
	asignArgsToParams(nodes[0].args,nodes[3].value)
	nodes[5].execute() # Funtion Body execution
#>
			;

	"Optional Function Parameters"
		params -> OPENBRACE paramList CLOSEBRACE <#
nodes[2].execute()
nodes[0].value=nodes[2].value
#>
			| <#
nodes[0].value=[]
#>
			;

	"One or more function parameter"
		paramList -> ID moreParams <#
nodes[2].init=[nodes[1].text]
nodes[2].execute()
nodes[0].value=nodes[2].value
#>
			;
	
	"Zero or more function parameters"
		moreParams -> COMMA ID moreParams <#
nodes[3].init=nodes[0].init+[nodes[2].text]
nodes[3].execute()
nodes[0].value=nodes[3].value
#>
			| <#
nodes[0].value=nodes[0].init
#>
			;

	"Function Body"
		functionBody -> functionStatement moreFunctionStatements <#
nodes[1].execute()
nodes[2].execute()
#>
			|
			;

	"Zero or more function statements"
		moreFunctionStatements ->  functionStatement moreFunctionStatements <#
nodes[1].execute()
nodes[2].execute()
#>
			|
			;

	"Function Statement"
		functionStatement -> return SEMICOLON <#
nodes[1].execute()
#>
			| print SEMICOLON <#
nodes[1].execute()
#>
			| if <#
nodes[1].execute()
#>
			| while <#
nodes[1].execute()
#>
			| _REIFY_ <#
#write("Trace> Reify code executed.\n") # NoComment=TraceExecution
nodes[1].execute()
#>
			| asignment SEMICOLON <#
nodes[1].execute()
#>
			| expression SEMICOLON <#
nodes[1].execute()
#write("Trace> Expression with value: "+str(nodes[1].value)+".\n") # NoComment=TraceExecution
#>
			;

	"Return Statement"
		return -> RETURN expression <#
nodes[2].execute()
global functionResult # Has the value to be returned in a function call
functionResult=nodes[2].value
#write("Trace> Function return: "+expressionToString(nodes[2].value)+"\n") # NoComment=TraceExecution
#>

			;

	"Print Statement"
		print -> PRINT OPENBRACE expression CLOSEBRACE <#
nodes[3].execute()
write(str(nodes[3].value)+"\n")
#>
			;

	"If Statement"
		if ->  IF expression OPENCURLYBRACE functionStatement moreFunctionStatements  CLOSECURLYBRACE else <#
nodes[2].execute()
if nodes[2].value:
	nodes[4].execute()
	nodes[5].execute()
else:
	nodes[7].execute()
#>
			;

	"Optional Else"
		else -> ELSE OPENCURLYBRACE functionStatement moreFunctionStatements  CLOSECURLYBRACE <#
nodes[3].execute()
nodes[4].execute()
#>
			| 
			;

	"While Statement"
		while ->  WHILE expression OPENCURLYBRACE functionStatement moreFunctionStatements  CLOSECURLYBRACE <#
nodes[2].execute()
while nodes[2].value:
	nodes[4].execute()
	nodes[5].execute()
	nodes[2].execute()
#>
			;

	"Import Statement"
		import ->  FROM ID IMPORT ID <#
#write("Trace> from "+nodes[2].text+" import "+nodes[4].text+"\n") # NoComment=TraceExecution
exec("global "+nodes[4].text+"\nfrom "+nodes[2].text+" import "+nodes[4].text+"\n")
#>
			| IMPORT ID <#
#write("Trace> import "+nodes[2].text+"\n") # NoComment=TraceExecution
exec("global "+nodes[2].text+"\nimport "+nodes[2].text+"\n")
#>
			;

	"Asignment Statement"
		asignment -> lvalue ASIGN expression <#
nodes[3].execute()
# "_rvalue" is the value of the expression's right-part
_rvalue=nodes[3].value
# Expression Asignment (the ID is going to be used in the rest of the execution,i.e. global)
exec("global "+getObjectName(nodes[1].text)+"\n"+nodes[1].text+"=_rvalue\n")
#write("Trace> Asignment statement: "+nodes[1].text+"="+expressionToString(nodes[3].value)+"\n") # NoComment=TraceExecution
#>
			;

	"Asignment Lvalue"
		lvalue -> ID attribute
			;

	"Lvalue Attribute Access"
		attribute -> DOT ID attribute
			|
			;

	"Method or function call"
		functionCall -> ID OPENBRACE args CLOSEBRACE <#
global functionResult # Has the value to be returned in a function call
functionResult=None
nodes[3].execute()
#write("Trace> Function call: "+nodes[1].text+"("+argsToText(nodes[3].value)+")\n") # NoComment=TraceExecution
# Builtin or user-defined function?
if not(functions.has_key(nodes[1].text)):
	# Builtin function
	nodes[0].value=invokeFunction(nodes[1].text,nodes[3].value)
else:
	# User-defined function
	function=functions[nodes[1].text]
	function.args=nodes[3].value # Args attribute has a the argument value list
	function.execute() # Function Call
	nodes[0].value=functionResult
#>
			| ID DOT ID OPENBRACE args CLOSEBRACE <#
global functionResult # Has the value to be returned in a function call
functionResult=None
nodes[5].execute()
#write("Trace> Method call: "+nodes[1].text+"."+nodes[3].text+"("+argsToText(nodes[5].value)+")\n") # NoComment=TraceExecution
# Is a builtin or a user-defined method?
exec("_temp=dir("+nodes[1].text+")\n")
try:
	_temp.index(nodes[3].text) # Exception raised if it is not a builtin function
	# A builtin method
	nodes[0].value=invokeFunction(nodes[1].text+"."+nodes[3].text,nodes[5].value)
except:
	# A user-defined method
	exec("_className="+nodes[1].text+".__class__.__name__\n") # Get the Class Name by using instrospection
	function=functions[_className+"."+nodes[3].text]
	function.args=nodes[5].value # Args attribute has a the argument value list
	exec("_object="+nodes[1].text+"\n") # Gets the implicit object
	nodes[5].value.insert(0,_object) # The implicit object is the firs argument in a method call
	function.execute() # Method Call
	nodes[0].value=functionResult
#>
			;

	"Function call Arguments"
		args -> expression moreArgs <#
nodes[1].execute()
nodes[2].init=[nodes[1].value]
nodes[2].execute()
nodes[0].value=nodes[2].value

#>
			| <#
nodes[0].value=[]
#>
			;

	"Zero or more function call arguments"
		moreArgs -> COMMA expression moreArgs <#
nodes[2].execute()
nodes[3].init=nodes[0].init+[nodes[2].value]
nodes[3].execute()
nodes[0].value=nodes[3].value
#>
			| <#
nodes[0].value=nodes[0].init
#>
			;

	"Binary Expression"
		expression -> term moreTerms  <#
# Every node has its evaluation on the attribute "value"
# Its initival value is saved on the attribute "init"
nodes[1].execute()
nodes[2].init=nodes[1].value
nodes[2].execute()
nodes[0].value=nodes[2].value
#>
			;

	"Zero or More Terms"
		moreTerms -> PLUS term moreTerms  <#
nodes[2].execute()
nodes[3].init=nodes[0].init+nodes[2].value
nodes[3].execute()
nodes[0].value=nodes[3].value
#>
			| MINUS term moreTerms <#
nodes[2].execute()
nodes[3].init=nodes[0].init-nodes[2].value
nodes[3].execute()
nodes[0].value=nodes[3].value
#>
			| <#
nodes[0].value=nodes[0].init
#>
			;

	"Binary Expression Term"
		term -> factor moreFactors <#
nodes[1].execute()
nodes[2].init=nodes[1].value
nodes[2].execute()
nodes[0].value=nodes[2].value
#>
			;

	"Zero or More Factors"
		moreFactors -> TIMES factor moreFactors <#
nodes[2].execute()
nodes[3].init=nodes[0].init*nodes[2].value
nodes[3].execute()
nodes[0].value=nodes[3].value
#>
			| BY factor moreFactors <#
nodes[2].execute()
nodes[3].init=nodes[0].init/nodes[2].value
nodes[3].execute()
nodes[0].value=nodes[3].value
#>
			| <#
nodes[0].value=nodes[0].init
#>
			;

	"Binary Expression Factor"
		factor -> comp moreComps <#
nodes[1].execute()
nodes[2].init=nodes[1].value
nodes[2].execute()
nodes[0].value=nodes[2].value
#>
			;

	"Zero or More Comparisons"
		moreComps -> LOWER comp moreComps <#
nodes[2].execute()
nodes[3].init=nodes[0].init<nodes[2].value
nodes[3].execute()
nodes[0].value=nodes[3].value
#>
			| LOWEREQUAL comp moreComps <#
nodes[2].execute()
nodes[3].init=nodes[0].init<=nodes[2].value
nodes[3].execute()
nodes[0].value=nodes[3].value
#>
			| GREATER comp moreComps <#
nodes[2].execute()
nodes[3].init=nodes[0].init>nodes[2].value
nodes[3].execute()
nodes[0].value=nodes[3].value
#>
			| GREATEREQUAL comp moreComps <#
nodes[2].execute()
nodes[3].init=nodes[0].init>=nodes[2].value
nodes[3].execute()
nodes[0].value=nodes[3].value
#>
			| EQUAL comp moreComps <#
nodes[2].execute()
nodes[3].init=nodes[0].init==nodes[2].value
nodes[3].execute()
nodes[0].value=nodes[3].value
#>
			| NOTEQUAL comp moreComps <#
nodes[2].execute()
nodes[3].init=nodes[0].init!=nodes[2].value
nodes[3].execute()
nodes[0].value=nodes[3].value
#>
			| <#
nodes[0].value=nodes[0].init
#>
			;

	"Binary Expression Comparison"
		comp -> 
			  FLOAT  <#
nodes[0].value=float(nodes[1].text)
#>
			| INTEGER  <#
nodes[0].value=int(nodes[1].text)
#>
			| OPENBRACE expression CLOSEBRACE <#
nodes[2].execute()
nodes[0].value=nodes[2].value
#>
			| MINUS expression <#
nodes[2].execute()
nodes[0].value=-nodes[2].value
#>
			| functionCall <#
nodes[1].execute()
nodes[0].value=nodes[1].value
#>
			| ID OPENBRACKET expression CLOSEBRACKET <#
nodes[3].execute()
# _index takes the Index-Expression value and _temp the whole expression value
_index=nodes[3].value
exec("_temp="+nodes[1].text+"[_index]\n")
nodes[0].value=_temp
#>
			| ID attribute <#
exec("_temp="+nodes[1].text+nodes[2].text+"\n")
nodes[0].value=_temp
#>
			| OPENBRACKET CLOSEBRACKET <#
nodes[0].value=[]
#>
			| OPENCURLYBRACE CLOSECURLYBRACE <#
nodes[0].value={}
#>
			| STRING  <#
nodes[0].value=nodes[1].text[1:len(nodes[1].text)-1]
#>
			;
}


Skip={  "\t";  "\n"; " "; }
NotSkip = {  }
