"Application Specification Module"

import string
import Tkinter # TK Interface
import tkMessageBox # ShowInfo

# ***************** Application Window **************************

class AppWindow:
    "Every application has its own graphic window"

    def __init__(self,title,rootWindow,application):
        "Application Windows Constructor"
        # Child Window
        self.master=Tkinter.Toplevel(rootWindow)
        self.master.protocol("WM_DELETE_WINDOW",self.close) 
        # Window title
        self.title=title
        self.master.wm_title(self.title)
        top=Tkinter.Frame(self.master)
        top.pack(fill=Tkinter.BOTH,expand=1)
        vbar=Tkinter.Scrollbar(top)
        # Text Control
        self.text = Tkinter.Text(top)
        self.text['font']=("courier new","10")
	# *** Added for PFC ***
	self.text['background'] = 'white'
	# *** End ***
        vbar["command"]=self.text.yview
        vbar.pack(side=Tkinter.RIGHT,fill=Tkinter.BOTH)
        self.text['yscrollcommand'] = vbar.set
        self.text.pack(side=Tkinter.TOP,fill=Tkinter.BOTH,expand=1)
        # Its application object
        self.application=application
        
    def close(self):
        "Method associated to the Window Close Event"
        if self.master.title()[-11:]=="- Running -":
            tkMessageBox.showinfo("nitrO Info","The application must be stopped to be closed (use the 'nitrO.stop()' method).")
        else:
            del self.application.nitrO.apps[self.application.name]
            self.master.destroy()
            
    def writeText(self,text):
        "Instead of printing text, this method must be used"
        self.text.insert(Tkinter.END,str(text))
        self.text.yview("moveto","1.0")

    def setTitle(self,title):
        "This method changes the window title; this may be known by asking its title attribute"
        self.title=title
        self.master.wm_title(self.title)

    def update(self):
        "Updates the window events"
        self.master.update()
    
# ************************** Application **************************

class Application:
    "Application Definition Class; its objects define an application and its language"
    
    def __init__(self,name,language,sourceCode,nitrO):
        "Application Constructor"
        # Application name
        self.name=name
        # Application Language
        self.language=language
        # Application-tree start node
        self.rootNode=0
        # Application Source Code
        self.sourceCode=sourceCode
        # Application Contexts (Runtime contexts)
        self.applicationGlobalContext={}
        # Tells if the application is running
        self.isRunning=0
        # Application window        
        self.window=AppWindow(self.describe(),nitrO.shell.master,self)
        # nitrO object
        self.nitrO=nitrO

    def describe(self):
        "Gives us information related to the Application at runtime"
        s="Application: "+self.name+". Language: " +self.language.name+"."
        if self.isRunning:
            s=s+" - Running -"
        else:
            s=s+" - Stopped -"
        return s


# ************************** TreeNode **************************

class TreeNode:
    "Application tree node. Associated with a language grammar nonTerminal object."
    
    def __init__(self,application,grammarSymbol,children,option,text):
        "TreeNode constructor"
        # Application being represented
        self.application=application
        # NonTerminal Grammar Symbol associated
        self.grammarSymbol=grammarSymbol
        # Its TreeNodes children
        self.children=children
        # The grammar-rule right part correct and its semantic actions
        self.option=option
        # Text being recognised by the symbol at runtime
        self.text=text

    def describe(self):
        "Application TreeNode description"
        s="Application TreeNode, representing an instance of the "+self.getName()+" grammar symbol with text '"+self.text+"' "
        if self.getType()!="Terminal" and len(self.option.actions)>0:
            s=s+"and actions: "
            for action in self.option.actions:
                s=s+action.action
        else:
            s=s+"\n"
        return s

    def describeWithChildren(self):
        "Application TreeNode description and its children short description"
        s=self.describe()
        if self.getType()!="Terminal":
            s=s+"\nChildren:\n"
            for child in self.children:
                s=s+"\t- "+child.describe()+"\n"
        return s

    def describeTree(self):
        "Whole Tree Description"
        s="\n************************************************************\n\n"
        s=s+self.describeWithChildren()
        for child in self.children:
            s=s+child.describeTree()
        return s        

    def getType(self):
        "Terminal or Not Terminal Symbol"
        className=self.grammarSymbol.__class__.__name__
        if className=="Terminal" or className=="Lambda":
            return "Terminal"
        else:
            return "NonTerminal"
                
    def getName(self):
        "Returns its grammar symbol name associated"
        if self.getType()=="Terminal":
            return '"'+self.grammarSymbol.value+'"'
        else:
            return "<"+self.grammarSymbol.name+">"

    def execute(self):
        "Runtime evaluation of the semantic actions associated"
        self.application.window.update()
        if self.getType()!="Terminal":
            # Default action (text concatenation)
            self.text=""
            for child in self.children:
                self.text=self.text+child.text
            # User actions
            for action in self.option.actions:
                action.execute(self)
        
# ******************** Module Testing Functions ********************

