"nitrO System Module"

import appSpec
import metaLang
import langSpec
import GI
import objectBrowser

import string
import Tkinter
import tkMessageBox
import code
import time
import tkFileDialog
import os

# ************************** Scanner ******************************
# Application Specification Grammar
# APPLICATION "=" STRING
# LANGUAGE "=" ( STRING | CODE ) ( "+" CODE )?

class Scanner:
    "Application Definition Scanner Class"

    # Keywords
    LANGUAGE=256
    APPLICATION=257
    # Tokens
    ASIGN=270
    STRING=271
    CODE=272
    PLUS=273

    # String Exception
    stringException="Language Specification Scanner Error"

    def __init__(self,string):
        "Constructor"
        # Last token
        self.token=0
        # Input String
        self.input=string
        # Buffer pointer
        self.charPos=0
        # Text Recognised
        self.text=""
        # Line number
        self.lineNumber=1

    def getChar(self):
        "Reads one Character"
        self.charPos=self.charPos+1
        char=self.input[self.charPos-1:self.charPos]
        if char=="\n":
            self.lineNumber=self.lineNumber+1
        return char

    def setChar(self):
        "Returns one Character to the Buffer"
        if self.input[self.charPos:self.charPos+1]=="\n":
            self.lineNumber=self.lineNumber-1
        self.charPos=self.charPos-1

    def error(self,msg=None):
        "Scanner Error"
        if msg:
            raise Scanner.stringException+" in line "+str(self.lineNumber)+".\n"+msg
        raise Scanner.stringException+" in line "+str(self.lineNumber)

    def getTheRest(self):
        "Finish Scanning and returns the rest"
        return self.input[self.charPos:]

    def getNextToken(self):
        "New token to be read"
        self.text=char=self.getChar()
        # Ignored Characters
        if char==" " or char=="\n" or char=="\t":
            self.token=self.getNextToken()
        # Character =
        elif char=="=":
            self.token=Scanner.ASIGN
        # Character +
        elif char=="+":
            self.token=Scanner.PLUS
        # Token STRING
        elif char=="\"":
            self.text=char=self.getChar()
            while char!="\"":
                char=self.getChar()
                if char!="\"":
                    self.text=self.text+char
            self.token=Scanner.STRING
        # Key Words
        elif char=="_" or (char>="a" and char<="z") or (char>="A" and char<="Z"):
            self.text=""
            while char=="_" or (char>="a" and char<="z") or (char>="A" and char<="Z"):
                self.text=self.text+char
                char=self.getChar()
            self.setChar()
            if string.upper(self.text)=="APPLICATION":
                self.token=Scanner.APPLICATION
            elif string.upper(self.text)=="LANGUAGE":
                self.token=Scanner.LANGUAGE
            else:
                self.error("Unknown Keyword.")
        # Token CODE
        elif char=="<":
            scanner=metaLang.Scanner(self.input[self.charPos-1:])
            if scanner.getNextToken()!=metaLang.Scanner.CODE:
                self.error("CODE token expected after the '<' character.")
            self.text=scanner.text
            self.charPos=self.charPos+scanner.charPos
            self.token=Scanner.CODE
        else:
            self.error("Unknown token.")
        return self.token

# ***************** MainWindow **************************

class MainWindow:
    "nitrO main window"
    def __init__(self,title,nitrO):
        # Master TK Window
        self.master=Tkinter.Tk()
        self.menu=Tkinter.Menu(self.master)
        self.master.protocol("WM_DELETE_WINDOW",self.quit)
        self.master.wm_title(title)
        self.master["menu"]=self.menu
        # Menu Items
        fileMenu=Tkinter.Menu(self.menu)
        self.menu.add_cascade(label="File",underline=0,menu=fileMenu)
        fileMenu.add_command(label="Open",underline=0,command=self.openFile)
        fileMenu.add_command(label="Clear console",underline=0,command=self.clearConsole)
        fileMenu.add_command(label="Clear output",underline=6,command=self.clearOutput)
        fileMenu.add_separator()
        fileMenu.add_command(label="Close",underline=1,command=self.quit)
        self.menu.add_command(label="Execute",underline=0,command=self.click)
        self.menu.add_command(label="Object Browser",underline=0,command=self.viewObjects)
        # Upper Frame
        top=Tkinter.Frame(self.master)
        top.pack(side=Tkinter.TOP,fill=Tkinter.BOTH,expand=1)
        # Lower Frame
        bottom=Tkinter.Frame(self.master)
        bottom.pack(side=Tkinter.TOP,fill=Tkinter.BOTH,expand=1)
        # Scroll Bars
        vbar1=Tkinter.Scrollbar(top)
        vbar2=Tkinter.Scrollbar(bottom)
        # Editing Text Controls
        self.text1 = Tkinter.Text(top,width=80,height=12)
        self.text2 = Tkinter.Text(bottom,width=80,height=12)
	# *** Added for PFC ***
        #self.text2["background"]="SystemScrollBar"
	self.text1["background"] = "white"
	self.text2["background"] = "gray"
	# *** End ***
        self.text2["state"]="disabled"
        self.text1['font']=self.text2['font']=("courier new","10")
        vbar1["command"]=self.text1.yview
        vbar2["command"]=self.text2.yview
        vbar1.pack(side=Tkinter.RIGHT, fill=Tkinter.Y)
        vbar2.pack(side=Tkinter.RIGHT, fill=Tkinter.Y)
        self.text1['yscrollcommand'] = vbar1.set
        self.text2['yscrollcommand'] = vbar2.set
        self.text1.pack(side=Tkinter.TOP,fill=Tkinter.BOTH,expand=1)
        self.text2.pack(side=Tkinter.TOP,fill=Tkinter.BOTH,expand=1)
        # Main nitrO object
        self.nitrO=nitrO
        # Context of Execution
        self.context={"write":self.write,"nitrO":self.nitrO}

    def quit(self):
        "Window Close Event associated method"
        if tkMessageBox.askokcancel("Quit","Do you really want to quit nitrO?"):
            self.master.destroy()

    def click(self):
        "Execution click event associated method"
        text=self.text1.get("0.0",Tkinter.END)
        exec(text,None,self.context)

    def clearConsole(self):
        "ClearConsole Event Method"
        self.text1.delete("0.0",Tkinter.END)

    def clearOutput(self):
        "ClearOutput Event Method"
        self.text2["state"]="normal"
        self.text2.delete("0.0",Tkinter.END)
        self.text2["state"]="disabled"

    def openFile(self):
        "OpenFile Event Method"
        fileName=tkFileDialog.askopenfilename(filetypes=[("nitrO script files", "*.ns"),("all files", "*",)],initialdir=self.nitrO.dir)
        if fileName:
            try:
                file=open(fileName)
                self.clearConsole()
                self.text1.insert(Tkinter.END,file.read())
            except IOError:
                tkMessageBox.showinfo("nitrO Info","The file can't be opened.")

    def viewObjects(self):
        "ObjectBrowser Event Method"
        objectBrowser.view("nitrO",self.nitrO,self.master)

    def write(self,text):
        "Method that writes a line at the end of the nitrO shell output. The code uses it as the 'write' function"
        self.text2["state"]="normal"
        self.text2.insert(Tkinter.END,"nitrO> "+str(text)+"\n")
        self.text2.yview("moveto","1.0")
        self.text2["state"]="disabled"

# ************************** NitrO ******************************

class NitrO:
    "Nitro Class; its unique instance 'nitrO' is the System's Facade (GOF)"

    def __init__(self):
        "NitrO Constructor"
        # Applications Dictionary
        self.apps={}
        # Source-Code Scanner (to create in the "loadApplication" method)
        self.scanner=0
        # Language Specification Directory
        #self.dir="c:/mis documentos/asignaturas/doctorado/tesis/metapython/"
        self.dir=os.getcwd()+os.sep
        # nitrO shell; created in the run method
        self.shell=MainWindow("nitrO shell",self)

    # String Exception
    stringException="Application or Language Identification Error"

    def error(self,msg=None):
        if msg:
            NitrO.stringException+" in line "+str(self.scanner.lineNumber)+".\n"+msg
        raise NitrO.stringException+" in line "+str(self.scanner.lineNumber)

    def describe(self):
        "Gives us information related to the nitrO System at runtime"
        s=""
        for key in self.apps.keys():
            s=s+"\t"+self.apps[key].describe()+"\n"
        if len(self.apps):
            return "There are "+str(len(self.apps))+" applications:\n"+s
        else:
            return "There are no applications."

    def loadApplication(self,code):
        "Loads the application code and language into the nitrO system; returns the app object"
        self.scanner=Scanner(code)
        token=self.scanner.getNextToken()
        if token!=Scanner.APPLICATION:
            self.error("APPLICATION token needed.")
        token=self.scanner.getNextToken()
        if token!=Scanner.ASIGN:
            self.error("= token needed.")
        token=self.scanner.getNextToken()
        if token!=Scanner.STRING:
            self.error("STRING token needed.")
        appName=self.scanner.text
        token=self.scanner.getNextToken()
        if token!=Scanner.LANGUAGE:
            self.error("LANGUAGE token needed.")
        token=self.scanner.getNextToken()
        if token!=Scanner.ASIGN:
            self.error("= token needed.")
        token=self.scanner.getNextToken()
        # Language Loading
        if token==Scanner.STRING:
            language=langSpec.Language(self.scanner.text)
            language=self.loadLanguage(language,languageName=self.scanner.text)
        elif token==Scanner.CODE:
            language=langSpec.Language("Nonamed")
            language=self.loadLanguage(language,languageSpec=self.scanner.text)
        else:
            self.error("Language definition needed.")
        # Any language definition added?
        appCode=self.scanner.getTheRest()
        try:
            # There might be a scanner error
            languageExt=self.scanner.getNextToken()==self.scanner.PLUS
        except:
            languageExt=0
        if languageExt:
            if self.scanner.getNextToken()!=Scanner.CODE:
                self.error("Language Extension Code needed after the + token.")
            languageExt=self.scanner.text
            appCode=self.scanner.getTheRest()
        application=self.apps[appName]=appSpec.Application(appName,language,appCode,nitrO)
        # If needed, a language specification module is added
        if languageExt:
            context={"nitrO":nitrO,"write":application.window.writeText,"application":application}
            exec(languageExt,context)
        return application

    def loadLanguage(self,language,languageName=None,languageSpec=None):
        """"Loads the language specification; it may be used a file (languageName) or the language specification (languageSpec).
        Returns the object-grammar language specification"""
        # Loading the language specification in a file
        if languageName:
            try:
                metaLanguageFile=open(self.dir+language.name+".ml")
            except IOError:
                raise language.name+" meta-language specification not available."
            languageSpec=metaLanguageFile.read()
        # Parsing
        metaLanguageParser=metaLang.Parser(languageSpec)
        metaLanguageParser.start()
        return metaLanguageParser.ogc.language

    def execute(self,code):
        "Executes an application in the nitrO system"
        # Taking the application and language object
        application=self.loadApplication(code)
        # Evaluates the application
        gi=GI.GI(application,self)
        nitrO.shell.write("Application being executed into the nitrO system.")
        gi.evaluate()

    def executeFile(self,fileName):
        "Executes an application specified in a file"
        try:
            applicationFile=open(fileName)
        except IOError:
            raise fileName+" application specification not available."
        applicationText=applicationFile.read()
        self.execute(applicationText)


    def run(self):
        "Executes the nitrO System"
        self.shell.master.mainloop()

    def stop(self,appID):
        "Stops an application that is running"
        if self.apps.has_key(appID):
            if self.apps[appID].isRunning:
                self.apps[appID].window.setTitle("Application: "+appID+". Language: " +self.apps[appID].language.name+". - Trying to Stop -")
                self.apps[appID].isRunning=0
                result="Trying to stop the "+appID+" application."
            else:
                result="The application "+appID+" is stopped."
        else:
            result="The application "+appID+" does not exists in the nitrO system."
        return result

# Global Nitro Object
nitrO=NitrO()
# Main loop
nitrO.run()

# ******************** Module Testing Functions ********************

# Scanner Testing Function
def testScanner():
    "Scanner Class Testing"
    s=""" Application = "My sample application"
          Language = "Python Language"
print "Python Code" # Python Comment
if a:
    print "More Python Code"
            """
    obj=Scanner(s)
    for i in range(6):
        token=obj.getNextToken()
        print "Token number:\t",obj.token,"\t",
        print "Token text:\t",obj.text,"\t",
        print "Line: ",obj.lineNumber
    print "\nThe code:\n"+obj.getTheRest()

# NitrO Testing Function
def testNitrO():
    "NitrO Class Testing"
    s=""" Application = "My sample application"
          Language = "Python Language"
print "Python Code" # Python Comment
if a:
    print "More Python Code"
            """
    ss=""" Application = "Another Application"
          Language = "Expression"
a=33+434-35
b=68-(a*3%2)
            """
    nitrO.loadApplication(s)
    nitrO.loadApplication(ss)
    # An application may be overwritten
    nitrO.loadApplication(s)
    print nitrO.describe()
    for key in nitrO.apps.keys():
        print "Source Code of \""+key+"\":" + nitrO.apps[key].sourceCode


def testLanguageSpec():
    "Load a language specification from disk an describes it and its appplication"
    s=""" Application = "My sample application"
          Language = "Expression"
print "Python Code" # Python Comment
if a:
    print "More Python Code"
            """
    nitrO.execute(s)
    print nitrO.apps["My sample application"].describe()
    print nitrO.apps["My sample application"].language.describe()
    # Language not defined
    nitrO.execute("Application=\"Don't know\" Language=\"Unknown\" MyCode")


# Application runned in its own thread testing function
from threading import Thread
class GenericThread(Thread):
    def __init__(self,method,param):
        self.method=method
        self.param=param
        Thread.__init__(self)

    def run(self):
        if self.param==None:
            self.method()
        else:
            self.method(self.param)

def testRun():
    "Application runned in its own thread testing function"
    code=""" Application = "My sample Calc application"
          Language = "Calc"
    FOR 1000
        CALCULATE 100+10
            """
    gt=GenericThread(nitrO.execute,code)
    gt.start()
    op="-"
    from time import sleep
    sleep(1)
    nitrO.apps["My sample Calc application"].language.syntacticSpec["statement"].options[0].actions[0].action="write( str(int(nodes[2].text)"+op+"int(nodes[4].text))+' ' )"
    while 1:
        sleep(1)

# Language description in the application specification testing function
def testLangSpec():
    "Language description in the application specification testing function"
    code=""" Application = "My sample Calc application"
          Language = <#
            Language=Calc
            Scanner={
                "Digit Token"
                digit -> "0" | "1" | "2" | "3" | "4" | "5" | "6" | "7" | "8" | "9"
		    ;
                "Number Token"
                NUMBER -> digit moreDigits
		    ;
		"Zero or more digits token"
		moreDigits -> digit moreDigits
		    |
		    ;
		"CALCULATE Key Word"
		CALCULATE -> "CALCULATE"
		    ;
		"FOR Key Word"
		FOR -> "FOR"
		    ;
		"PLUS token"
		PLUS -> "+"
		    ;
                }
            Parser = {
                "Initial Free-Context Rule"
                S -> whileStatement <#
nodes[1].execute()
#>
			;
		   "While Statement"
		   whileStatement -> FOR NUMBER statement <#
for i in range(int(nodes[2].text)):
	nodes[3].execute()
#>
                   ;
		    "Language statements"
                statement -> CALCULATE NUMBER PLUS NUMBER <#
write( str(int(nodes[2].text)+int(nodes[4].text))+' ' )
#>
                   ;
                }
            Skip={  "\t";  "\n"; " ";}
            NotSkip = {  }
          #>

    FOR 1000
        CALCULATE 100+10
            """
    nitrO.execute(code)

# Testing of applications located in files
def testAppFile():
    "Testing of applications located in files"
    nitrO.executeFile(nitrO.dir+"calcApp1.na")
    nitrO.executeFile(nitrO.dir+"calcApp2.na")
    nitrO.shell.write(nitrO.describe())
