# -*- coding: ISO-8859-1 -*-

import sys, operator, guidgen

class JClass(object):
	def __init__(self, name, parent = None):
		self.__name = name
		self.__parent = parent
		self.__isA = {name: self}
		self.__fields = {}
		self.__ownFields = {}
		self.__methods = {}
		self.__ownMethods = {}
		self.__methodGroups = {}
		self.__constructors = JMethodGroup(name)
		if parent:
			self.__isA.update(parent.__isA)
			self.__fields.update(parent.__fields)
			self.__methods.update(parent.__methods)
			for name, group in parent.__methodGroups.items():
				self.__methodGroups[name] = group.clone()

	def __repr__(self): return '<Java-- class %s>' % self.__name

	def isNullClass(self): return False

	def getName(self): return self.__name
	def getParent(self): return self.__parent
	def getFields(self): return self.__fields
	def getField(self, name):
		try:
			return self.__fields[name]
		except KeyError, e:
			raise 'La clase %s no tiene ningn campo llamado %s' \
				% (self, name)

	def updateFromParent(self):
		parent = self.__parent
		if parent:
			self.__fields.update(parent.__fields)
			self.__methods.update(parent.__methods)
			for name, group in parent.__methodGroups.items():
				self.__methodGroups[name] = group.clone()

	def isA(self, klass):
		name = klass.getName()
		return self.__isA.has_key(name)

	def addFields(self, fields):
		for f in fields: self.addField(f)

	def addField(self, field):
		name = field.getName()
		if self.__ownFields.has_key(name):
			raise 'Ya existe un campo llamado %s' % name
		self.__fields[name] = field
		self.__ownFields[name] = field

	def addMethods(self, methods):
		for m in methods: self.addMethod(m)

	def addMethod(self, method):
		method.setClass(self)
		mangledName = method.getMangledName()
		if self.__ownMethods.has_key(mangledName):
			raise 'Redefinido el mtodo "%s" de la clase %s' \
				% (method, self.getName)
		try:
			super = self.__methods[mangledName]
			if method.getReturnType() != super.getReturnType():
				raise 'Si se sobreescribe un mtodo, los' \
					+ ' tipos de retorno deben coincidir'
		except KeyError, e:
			super = None
		method.setSuperMethod(super)
		self.__methods[mangledName] = method
		self.__ownMethods[mangledName] = method
		name = method.getName()
		try:
			group = self.__methodGroups[name]
		except KeyError, e:
			group = JMethodGroup(name)
			self.__methodGroups[name] = group
		group.addMethod(method)

	def addConstructors(self, constructors):
		for c in constructors: self.addConstructor(c)

	def addConstructor(self, constructor):
		constructor.setClass(self)
		self.__constructors.addMethod(constructor)

	# Aade el constructor por defecto si no se defini ningn constructor
	def addDefaultConstructor(self):
		def emptyCode(interpreter, self, instance, args): pass
		if self.__constructors.isEmpty():
			self.addConstructor(
				JConstructor(self, (),
					emptyCode, False))

	# Devuelve una lista con los mtodos que se podran lanzar si se llama
	# al mtodo name de la clase con los argumentos de los tipos indicados
	# en argTypes (si la lista tiene longitud > 1 hay ambigedad, y si es
	# 0 es porque ninguno de los mtodos es aplicable)
	def getMethods(self, name, argTypes):
		try:
			return self.__methodGroups[name].match(argTypes)
		except KeyError, e:
			raise 'La clase %s no tiene ningn mtodo llamado %s' \
				% (self, name)

	# Devuelve el mtodo que hay que llamar con los argumentos dados
	def getMethod(self, name, argTypes):
		methods = self.getMethods(name, argTypes)
		numMethods = len(methods)
		if numMethods < 1:
			raise 'No se puede llamar al mtodo %s con parmetros %s' \
				% (name, argTypes)
		elif numMethods > 1:
			candidates = "\n".join([str(i) for i in methods])
			raise ("Ambigedad en la llamada al mtodo %s" \
					+ " con parmetros %s\n" \
					+ "Candidatos:\n%s") \
				% (name, argTypes, candidates)
		else:
			return methods[0]

	def getMethodByMangledName(self, mangledName):
		try:
			return self.__methods[mangledName]
		except KeyError, e:
			raise 'La clase %s no tiene ningn mtodo' \
					+ ' con mangled name %s' \
				% (self, mangledName)

	def getConstructors(self, argTypes):
		return self.__constructors.match(argTypes)

	def getConstructor(self, argTypes):
		constructors = self.getConstructors(argTypes)
		numConstructors = len(constructors)
		if numConstructors < 1:
			raise ('No hay ningn constructor de %s que admita' \
					+ ' los argumentos de tipos %s') \
				% (self.getName(), argTypes)
		elif numConstructors > 1:
			candidates = "\n".join([str(i) for i in constructors])
			raise "Ambigedad en la llamada al constructor de %s" \
					+ " con parmetros %s\n" \
					+ "Candidatos:\n%s" \
				% (self.getName(), argTypes, candidates)
		else:
			return constructors[0]

	def getDefaultConstructor(self):
		return self.getConstructor(())

	# Devuelve la distancia entre esta clase y una supuesta clase base
	# suya, o -1 si ancestor no es una clase base de sta
	def distance(self, ancestor):
		def calcDistance(fromClass, ancestor):
			if ancestor.__name == fromClass.__name: return 0
			else: return 1 + calcDistance(fromClass.__parent, ancestor)
		if self.isNullClass(): return 0
		elif not self.isA(ancestor): return -1
		else: return calcDistance(self, ancestor)

	#def invokeByMangledName(self, interpreter, mangledName, instance, args):
	#	method = self.getMethodByMangledName(mangledName)
	#	return m.invoke(interpreter, instance, args)

	def newInstance(self, guid = None):
		inst = JInstance(self, guid)
		return inst

	def createFields(self, owner):
		return self.__createFields({}, owner)

	def __createFields(self, fields, owner):
		parent = self.__parent
		if parent:
			fields = parent.__createFields(fields, owner)
			parentFields = fields[parent]
			myFields = parentFields.copy()
		else:
			myFields = {}

		# El siguiente bucle puede provocar que el programa entre
		# en un bucle infinito y rompa (caso de tener p.ej. una
		# clase A con un miembro de tipo B y una clase B con un
		# miembro de tipo A: habra que comprobar previamente que
		# no se dan casos de dependencia mutua como se)
		for name, f in self.__ownFields.items():
			klass = f.getClass()
			inst = f.getInit().getInstance()
			myFields[name] = JRef(klass, inst, owner)

		fields[self] = myFields
		return fields

# ----------------------------------------------------------------------------

# Clase que representa al tipo nulo, es un caso especial por tener algunos
# comportamientos diferentes a las de una clase tpica (concretamente el isA)
class JNullClass(JClass):
	def __init__(self): JClass.__init__(self, '<null>')
	def isA(self, klass): return True
	def isNullClass(self): return True

# ----------------------------------------------------------------------------

# Grupo de mtodos con el mismo nombre pero diferentes parmetros

class JMethodGroup(object):
	def __init__(self, name):
		self.__name = name
		self.__methods = {}

	def __repr__(self):
		return '<Java-- method group %s: %s>' \
			% (self.__name, self.__methods)

	def addMethod(self, method):
		self.__methods[method.getMangledName()] = method

	def getMethodByMangledName(self, name): return self.__methods[name]

	def match(self, argTypes):
		# Calculamos la distancia de cada mtodo 
		pairs = [(x.distance(argTypes), x)
			for x in self.__methods.values()]
		# Quitamos las distancias menores que 0 (mtodos no aplicables)
		pairs = [x for x in pairs if x[0] >= 0]
		# Si no nos quedan mtodos podemos cortar ya
		if not pairs: return []
		# Calculamos la distancia mnima
		distances = [x[0] for x in pairs]
		minDist = reduce(min, distances)
		# Descartamos los mtodos que no estn a la distancia mnima y
		# los que nos queden son los que podran servir
		return [x[1] for x in pairs if x[0] == minDist]

	def isEmpty(self): return not self.__methods

	def clone(self):
		theClone = JMethodGroup(self.__name)
		theClone.__methods.update(self.__methods)
		return theClone

# ----------------------------------------------------------------------------

class JMethod(object):
	def __init__(self, name, returnType, args, code = None):
		self.__name = name
		self.__returnType = returnType
		# args es una secuencia de pares (tipo, nombre),
		self.__args = args
		# Cogemos slo los tipos
		self.__argTypes = [x[1] for x in args]
		# Ahora nos quedamos con los nombres de los tipos
		self.__argTypeNames = [x.getName() for x in self.__argTypes]
		# Y por ltimo guardamos los nombres de los argumentos
		self.__argNames = [x[0] for x in args]
		self.__mangledName = JMethod.mangle(name, self.__argTypeNames)
		self.__code = code

	def mangle(name, typeNames): return '%s@%s' % (name, '&'.join(typeNames))
	mangle = staticmethod(mangle)

	def __repr__(self):
		return '<Java-- method "%s %s.%s(%s)">' % (
			self.__returnType.getName(), self.__klass.getName(),
			self.__name, ', '.join(self.__argTypeNames))

	def getName(self): return self.__name
	def getMangledName(self): return self.__mangledName
	def getClass(self): return self.__klass
	def setClass(self, klass): self.__klass = klass
	def getReturnType(self): return self.__returnType
	def getArgs(self): return self.__args
	def getSuperMethod(self): return self.__superMethod
	def setSuperMethod(self, super): self.__superMethod = super
	def getArgTypes(self): return self.__argTypes
	def getArgTypeNames(self): return self.__argTypeNames

	# La distancia del mtodo a los argumentos ser el mximo de las
	# distancias de cada argumento a los de la definicin del mtodo
	def distance(self, args):
		# Si la longitud de la lista de argumentos que nos pasan
		# no coincide con la que tenemos que recibir, no valemos
		if len(args) != len(self.__argTypes): return -1
		# Si coinciden y args == [], entonces distancia cero
		if not args: return 0
		# Calculamos la distancia de cada argumento a los que
		# realmente nos pasan...
		distances = [x.distance(y)
			for x, y in zip(args, self.__argTypes)]
		# ... luego calculamos el mximo y el mnimo...
		maxDist = reduce(max, distances)
		minDist = reduce(min, distances)
		# ... y retornamos el mximo EXCEPTO si el mnimo es < 0
		# (que slo suceder si algn parmetro no es convertible)
		if minDist < 0: return -1
		else: return maxDist

	def setCode(self, code): self.__code = code

	def invoke(self, interpreter, instance, args):
		return self.__code(interpreter, self, instance, args)

# ----------------------------------------------------------------------------

# Para la ejecucin de los return lanzamos una excepcin de este tipo, y as
# nos ahorramos tener que hacer comprobaciones constantemente
class JReturn:
	def __init__(self, value):
		self.__value = value

	def getValue(self):
		return self.__value

# Envoltorio para poder llamar a cdigo del usuario como si fuese una funcin
# primitiva
class UserCodeWrapper(object):
	def __init__(self, code):
		self.__code = code

	def __call__(self, interpreter, method, instance, args):
		symbolTable = interpreter.getSymbolTable()
		symbolTable.methodEnter(instance, method, args)
		try:
			result = interpreter.parse(self.__code)
			if result:
				result = result.getInstance()
			else:
				result = voidInstance
				#if self.getReturnType() != voidClass:
					# Deberamos alertar de que se termin
					# el mtodo sin encontrar un return
		except JReturn, r:
			result = r.getValue()
		symbolTable.methodExit()
		return result

# ----------------------------------------------------------------------------

class JConstructor(JMethod):
	def __init__(self, klass, args, code = None,
		hasSuperStatement = False):
		name = klass.getName()
		JMethod.__init__(self, klass.getName(), voidInstance,
			args, code)
		self.__hasSuperStatement = hasSuperStatement

	def __repr__(self):
		return '<Java-- constructor "%s.%s(%s)">' % (
			self.getClass().getName(), self.getName(),
			', '.join(self.getArgTypeNames()))

	def setHasSuperStatement(self, hasSuperStatement):
		self.__hasSuperStatement = hasSuperStatement

	def invoke(self, interpreter, instance, args):
		parent = self.getClass().getParent()
		if parent and not self.__hasSuperStatement:
			parentCtor = parent.getDefaultConstructor()
			parentCtor.invoke(interpreter, instance, args)
		JMethod.invoke(self, interpreter, instance, args)

# ----------------------------------------------------------------------------

class JField(object):
	def __init__(self, name, klass, init):
		self.__name = name
		self.__klass = klass
		self.__init = init

	def getName(self): return self.__name
	def getClass(self): return self.__klass
	def getInit(self): return self.__init

# ----------------------------------------------------------------------------

class JInstance(object):
	def __init__(self, klass, guid = None):
		self.klass = klass
		self.value = None

		# Para la persistencia, cada instancia tendr un GUID
		if guid is None:
			self.guid = guidgen.guid()
		else:
			self.guid = str(guid)

		# No nos estamos guardando (para la recursividad en store())
		self.storing = False

		# Al crearnos estamos modificndonos
		self.modified = True

		# Por defecto somos almacenables y no persistentes
		self.nonStorable = False
		self.persistent = False

		self.fields = klass.createFields(self)

	def getID(self): return self.guid
		
	def isNonStorable(self): return self.nonStorable
	def setNonStorable(self): self.nonStorable = True

	def __getstate__(self):
		dict = self.__dict__.copy()
		del dict['storing']
		del dict['modified']
		del dict['persistent']
		dict['klass'] = self.klass.getName()
		fields = {}
		for (klass, fieldDict) in self.fields.items():
			fields[klass.getName()] = fieldDict
		dict['fields'] = fields
		if dict.has_key('manager'): del dict['manager']
		if hasattr(self.klass, 'extra_getstate'):
			self.klass.extra_getstate(self, dict)
		return dict

	def __setstate__(self, dict):
		self.__dict__.update(dict)

		self.storing = False
		self.modified = False
		self.persistent = False
		self.not_ready = True

	def isA(self, klass): return self.klass.isA(klass)
	def getClass(self): return self.klass
	def getField(self, klass, field): return self.fields[klass][field]
	def setField(self, klass, field, ref):
		self.fields[klass][field].setInstanceFrom(ref)
	def getFields(self, klass): return self.fields[klass] 
	def getValue(self): return self.value
	def setValue(self, value):
		self.value = value
		self.setModified()
	def isNull(self): return self.klass.isNullClass()
	def isModified(self): return self.modified
	def setModified(self, modified = True):
		self.modified = modified

		# Si somos persistentes, debemos avisar al Persistence Manager
		# de que nos hemos modificado
		if self.isPersistent():
			self.manager.notifyModified(self, modified)

	def isPersistent(self): return self.persistent
	def makePersistent(self, manager):
		if not (self.isNonStorable() or self.isPersistent()):
			self.persistent = True
			self.manager = manager
			refs = self.getRefs()
			for ref in refs.values():
				inst = ref.getInstance()
				inst.makePersistent(manager)
			if hasattr(self.klass, 'extra_makePersistent'):
				self.klass.extra_makePersistent(self, manager)
	def makeTransient(self):
		if not self.isPersistent(): return
		self.persistent = False
		refs = self.getRefs()
		for ref in refs.values():
			inst = ref.getInstance()
			inst.makeTransient()
			if hasattr(self.klass, 'extra_makeTransient'):
				self.klass.extra_makeTransient(self, manager)

	def store(self, storage):
		# Para cortar el proceso recursivo
		if self.storing: return
		self.storing = True

		# Primero nuestros miembros
		for (klass, fieldDict) in self.fields.items():
			for ref in fieldDict.values():
				inst = ref.getInstance()
				inst.store(storage)

		# Si nos tenemos que guardar, nos guardamos
		if not self.isNonStorable() and (self.modified \
			or not storage.has_key(self.getID())):
			storage.storeObject(self)
			self.modified = False

		# Para casos ms complejos, como arrays y diccionarios
		if hasattr(self.klass, 'extra_store'):
			self.klass.extra_store(self, storage)

		self.storing = False
		return self.getID()

	def getRefs(self):
		refs = {}
		for fieldDict in self.fields.values():
			# Un simple update del diccionario no nos vale,
			# porque queremos todas las referencias a miembros,
			# y en alguna clase se pudo ocultar alguno de los de
			# su clase base dndole el mismo nombre...
			#refs.update(fieldDict)

			for ref in fieldDict.values():
				refs[ref] = ref
		return refs

	def restore(self, interpreter, storage):
		if hasattr(self, 'not_ready'):
			# Restauramos la clase
			self.klass = interpreter.getClassByName(self.klass)

			# Restauramos los campos
			fields = {}
			for (klass, fieldDict) in self.fields.items():
				klass = interpreter.getClassByName(klass)
				fields[klass] = fieldDict
			self.fields = fields

			# Restauramos las referencias
			refs = self.getRefs()
			for ref in refs.values():
				ref.restore(interpreter, storage)

			# Para casos ms complejos, como arrays y diccionarios
			if hasattr(self.klass, 'extra_restore'):
				self.klass.extra_restore(self, interpreter,
							storage)

			del self.not_ready

# ----------------------------------------------------------------------------

class JRef(object):
	def __init__(self, klass, instance, owner = None):
		self.klass = klass
		self.owner = owner
		self.setInstance(instance)

	def getClass(self): return self.klass
	def getInstance(self): return self.instance
	def isNull(self): return self.instance.isNull()

	def setOwner(self, owner):
		self.owner = owner

	def setInstance(self, instance):
		self.instance = instance
		if self.owner:
			if self.owner.isPersistent():
				self.owner.manager.makePersistent(instance)
			self.owner.setModified(self)

	def setInstanceFrom(self, ref):
		self.setInstance(ref.getInstance())

	def __getstate__(self):
		dict = self.__dict__.copy()
		dict['klass'] = self.klass.getName()
		if self.owner:
			dict['owner'] = self.owner.getID()
		dict['instance'] = self.instance.getID()
		return dict

	def __setstate__(self, dict):
		self.__dict__.update(dict)
		self.not_ready = True

	def restore(self, interpreter, storage):
		if hasattr(self, 'not_ready'):
			self.klass = interpreter.getClassByName(self.klass)
			if self.owner:
				self.owner = storage.retrieveObject(self.owner)
			self.instance = storage.retrieveObject(self.instance)
			del self.not_ready

#-----------------------------------------------------------------------------

nullClass = JNullClass()
nullInstance = nullClass.newInstance(guid = 0) # Similar al None de Python...
# Null y similares siempre sern las mismas para todo programa
nullInstance.setNonStorable()

voidClass = JClass('<void>')
voidInstance = voidClass.newInstance(guid = 1)
voidInstance.setNonStorable()

classClass = JClass('<class>')

boolClass = JClass('Bool')
boolInstance = boolClass.newInstance(guid = 2) # Similar al T de Lisp
boolInstance.setNonStorable()

objectClass = JClass('Object')
objectClass.addDefaultConstructor()

integerClass = JClass('Integer', objectClass)
stringClass = JClass('String', objectClass)
consoleClass = JClass('Console', objectClass)
arrayClass = JClass('Array', objectClass)
dictClass = JClass('Dictionary', objectClass)
managerClass = JClass('PersistenceManager', objectClass)

primitiveClasses = [
	nullClass,
	voidClass,
	classClass,
	boolClass,
	objectClass,
	integerClass,
	stringClass,
	consoleClass,
	arrayClass,
	dictClass,
	managerClass
]
nonStorableInstances = {
	nullInstance.getID(): nullInstance,
	voidInstance.getID(): voidInstance,
	boolInstance.getID(): boolInstance
}

#-----------------------------------------------------------------------------

# Utilidad para crear booleanos rpidamente

def makeBool(test):
	if test: bb = boolInstance
	else: bb = nullInstance
	return bb

# Para tipos simples como cadenas y enteros

def makeSimple(klass, value):
	inst = klass.newInstance()
	inst.setValue(value)
	return inst

#-----------------------------------------------------------------------------

# Funciones auxiliares para crear operadores

def makeBinaryAOp(klass, op): # Operadores aritmticos binarios
	def f(interpreter, method, instance, args):
		arg = args[0].getInstance()
		if arg.isNull(): raise 'El argumento del operador es nulo'
		result = klass.newInstance()
		result.setValue(op(instance.getValue(), arg.getValue()))
		return result
	return f

def makeBinaryLOp(op): # Operadores lgicos binarios
	def f(interpreter, method, instance, args):
		arg = args[0].getInstance()
		if arg.isNull(): raise 'El argumento del operador es nulo'
		return makeBool(op(instance.getValue(), arg.getValue()))
	return f

#-----------------------------------------------------------------------------

# Mtodos de la clase Object

objectClass.updateFromParent()

def objectMethod_id(interpreter, method, instance, args):
	return makeSimple(stringClass, instance.getID())
def objectMethod_makePersistent(interpreter, method, instance, args):
	manager = interpreter.getPersistenceManager()
	manager.makePersistent(instance)
	return voidInstance
def objectMethod_makeTransient(interpreter, method, instance, args):
	manager = interpreter.getPersistenceManager()
	manager.makeTransient(instance)
	return voidInstance

objectClass.addMethods([
		JMethod('id', stringClass, (), objectMethod_id),
		JMethod('makePersistent', voidClass, (),
			objectMethod_makePersistent),
		JMethod('makeTransient', voidClass, (),
			objectMethod_makeTransient)
	])

#-----------------------------------------------------------------------------

# Mtodos de la clase Bool

boolClass.updateFromParent()

# Constructores
#def boolCtor_default(interpreter, ctor, instance, args):
#	raise 'No se permite construir instancias de la clase Bool'

#boolClass.addConstructors([
#		JConstructor(boolClass, (),
#			boolCtor_default, False)
#	])

#-----------------------------------------------------------------------------

# Mtodos de la clase Integer

integerClass.updateFromParent()

# Operadores aritmticos
integerMethod_sum = makeBinaryAOp(integerClass, operator.add)
integerMethod_sub = makeBinaryAOp(integerClass, operator.sub)
integerMethod_mul = makeBinaryAOp(integerClass, operator.mul)
integerMethod_div = makeBinaryAOp(integerClass, operator.div)
integerMethod_mod = makeBinaryAOp(integerClass, operator.mod)
def integerMethod_preinc(interpreter, method, instance, args):
	instance.setValue(instance.getValue() + 1)
	return instance
def integerMethod_postinc(interpreter, method, instance, args):
	value = instance.getValue()
	instance.setValue(value + 1)
	return makeSimple(integerClass, value)
def integerMethod_predec(interpreter, method, instance, args):
	instance.setValue(instance.getValue() - 1)
	return instance
def integerMethod_postdec(interpreter, method, instance, args):
	value = instance.getValue()
	instance.setValue(value - 1)
	return makeSimple(integerClass, value)

# Operadores de comparacin
integerMethod_lt = makeBinaryLOp(operator.lt)
integerMethod_le = makeBinaryLOp(operator.le)
integerMethod_gt = makeBinaryLOp(operator.gt)
integerMethod_ge = makeBinaryLOp(operator.ge)
integerMethod_eq = makeBinaryLOp(operator.eq)
integerMethod_ne = makeBinaryLOp(operator.ne)
	
# Otros mtodos
def integerMethod_set(interpreter, method, instance, args):
	arg = args[0].getInstance()
	if arg.isNull():
		raise 'El parmetro de set() no puede ser nulo'
	instance.setValue(arg.getValue())
	return voidInstance()
def integerMethod_toString(interpreter, method, instance, args):
	return makeSimple(stringClass, str(instance.getValue()))

integerClass.addMethods([
		JMethod('operator+', integerClass,
			(('x', integerClass),), integerMethod_sum),
		JMethod('operator-', integerClass,
			(('x', integerClass),), integerMethod_sub),
		JMethod('operator*', integerClass,
			(('x', integerClass),), integerMethod_mul),
		JMethod('operator/', integerClass,
			(('x', integerClass),), integerMethod_div),
		JMethod('operator%', integerClass,
			(('x', integerClass),), integerMethod_mod),
		JMethod('operator++', integerClass, (),
			integerMethod_preinc),
		JMethod('operator++', integerClass,
			(('x', integerClass),), integerMethod_postinc),
		JMethod('operator--', integerClass, (),
			integerMethod_predec),
		JMethod('operator--', integerClass,
			(('x', integerClass),), integerMethod_postdec),
		JMethod('operator<', boolClass,
			(('x', integerClass),), integerMethod_lt),
		JMethod('operator<=', boolClass,
			(('x', integerClass),), integerMethod_le),
		JMethod('operator>', boolClass,
			(('x', integerClass),), integerMethod_gt),
		JMethod('operator>=', boolClass,
			(('x', integerClass),), integerMethod_ge),
		JMethod('operator==', boolClass,
			(('x', integerClass),), integerMethod_eq),
		JMethod('operator!=', boolClass,
			(('x', integerClass),), integerMethod_ne),
		JMethod('set', voidClass,
			(('x', integerClass),), integerMethod_set),
		JMethod('toString', stringClass, (),
			integerMethod_toString)
	])

# Constructores

def integerCtor_default(interpreter, ctor, instance, args):
	instance.setValue(0)
	return voidInstance

def integerCtor_int(interpreter, ctor, instance, args):
	arg = args[0].getInstance()
	if arg.isNull(): raise 'El argumento del constructor es nulo'
	instance.setValue(arg.getValue())
	return voidInstance

integerClass.addConstructors([
		JConstructor(integerClass, (),
			integerCtor_default, False),
		JConstructor(integerClass,
			(('x', integerClass),),
			integerCtor_int, False)
	])

#-----------------------------------------------------------------------------

# Mtodos de la clase String

stringClass.updateFromParent()

# Operador de concatenacin
stringMethod_sum = makeBinaryAOp(stringClass, operator.add)

# Otros mtodos
def stringMethod_set(interpreter, method, instance, args):
	arg = args[0].getInstance()
	if arg.isNull():
		raise 'El parmetro de set() no puede ser nulo'
	instance.setValue(arg.getValue())
	return voidInstance()

stringClass.addMethods([
		JMethod('operator+', stringClass,
			(('x', stringClass),), stringMethod_sum),
		JMethod('set', voidClass,
			(('x', stringClass),), stringMethod_set)
	])

# Constructores

def stringCtor_default(interpreter, ctor, instance, args):
	instance.setValue('')
	return voidInstance

def stringCtor_str(interpreter, ctor, instance, args):
	arg = args[0].getInstance()
	if arg.isNull(): raise 'El argumento del constructor es nulo'
	instance.setValue(arg.getValue())
	return voidInstance

stringClass.addConstructors([
		JConstructor(stringClass, (),
			stringCtor_default, False),
		JConstructor(stringClass,
			(('x', stringClass),),
			stringCtor_str, False)
	])

#-----------------------------------------------------------------------------

# Mtodos de la clase Console

consoleClass.updateFromParent()

def consoleMethod_print(interpreter, method, instance, args):
	arg = args[0].getInstance()
	if arg.isNull():
		s = '<null>'
	else:
		s = arg.getValue()
	interpreter.write(s)
	return voidInstance

def consoleMethod_println(interpreter, method, instance, args):
	arg = args[0].getInstance()
	if arg.isNull():
		s = '<null>'
	else:
		s = arg.getValue()
	interpreter.write(s + "\n")
	return voidInstance

consoleClass.addMethods([
		JMethod('print', voidClass, (('x', stringClass),),
			consoleMethod_print),
		JMethod('println', voidClass, (('x', stringClass),),
			consoleMethod_println)
	])
consoleClass.addDefaultConstructor()

#-----------------------------------------------------------------------------

# Mtodos de la clase Array

def arrayClass_extraMakePersistent(instance, manager):
	list = instance.getValue()[0]
	for instance in list:
		instance.makePersistent(manager)

def arrayClass_extraGetstate(instance, dict):
	list, size = instance.getValue()
	list = [instance.getID() for instance in list]
	dict['value'] = (list, size)

def arrayClass_extraStore(instance, storage):
	list = instance.getValue()[0]
	for instance in list:
		instance.store(storage)

def arrayClass_extraRestore(instance, interpreter, storage):
	list = instance.getValue()[0]
	for inst in list:
		inst.restore(interpreter, storage)

arrayClass.extra_makePersistent = arrayClass_extraMakePersistent
arrayClass.extra_getstate = arrayClass_extraGetstate
arrayClass.extra_store = arrayClass_extraStore
arrayClass.extra_restore = arrayClass_extraRestore
arrayClass.updateFromParent()

def arrayMethod_getSize(interpreter, method, instance, args):
	return makeSimple(integerClass, instance.getValue()[1])

def arrayMethod_setSize(interpreter, method, instance, args):
	arg = args[0].getInstance()
	if arg.isNull():
		raise 'El argumento de setSize() no puede ser nulo'
	size = arg.getValue()
	if size < 0:
		raise 'El argumento de setSize() debe ser >= 0'
	list = [JRef(objectClass, nullInstance) for i in range(size)]
	oldlist, oldsize = instance.getValue()
	minsize = min(size, oldsize)
	list[0:minsize] = oldlist[0:minsize]
	instance.setValue((list, size))
	return voidInstance

def arrayMethod_getItem(interpreter, method, instance, args):
	pos = args[0].getInstance()
	if pos.isNull():
		raise 'La posicin para getItem() no puede ser nula'
	pos = pos.getValue()
	list, size = instance.getValue()
	if (pos < 0) or (pos >= size):
		raise ('ndice fuera de rango en getItem() ' \
			+ '(%d, vlido rango [0..%d])') \
				% (pos, size-1)
	return list[pos].getInstance()

def arrayMethod_setItem(interpreter, method, instance, args):
	pos = args[0].getInstance()
	if pos.isNull():
		raise 'La posicin para setItem() no puede ser nula'
	pos = pos.getValue()
	list, size = instance.getValue()
	if (pos < 0) or (pos >= size):
		raise ('ndice fuera de rango en setItem() ' \
			+ '(%d, vlido rango [0..%d])') \
				% (pos, size-1)
	item = args[1].getInstance()
	list[pos].setInstance(item)
	if instance.isPersistent():
		instance.manager.makePersistent(item)
	instance.setModified()
	return voidInstance

arrayClass.addMethods([
		JMethod('getSize', integerClass, (),
			arrayMethod_getSize),
		JMethod('setSize', voidClass, (('size', integerClass),),
			arrayMethod_setSize),
		JMethod('getItem', objectClass, (('index', integerClass),),
			arrayMethod_getItem),
		JMethod('setItem', voidClass,
			(('index', integerClass), ('item', objectClass),),
			arrayMethod_setItem),
	])

# Constructores

def arrayCtor_default(interpreter, ctor, instance, args):
	#size = 1
	#list = [nullInstance]
	size = 0
	list = []
	instance.setValue((list, size))
	return voidInstance

def arrayCtor_int(interpreter, ctor, instance, args):
	arg = args[0].getInstance()
	if arg.isNull(): raise 'El argumento del constructor es nulo'
	size = arg.getValue()
	list = [JRef(objectClass, nullInstance) for i in range(size)]
	instance.setValue((list, size))
	return voidInstance

arrayClass.addConstructors([
		JConstructor(arrayClass, (),
			arrayCtor_default, False),
		JConstructor(arrayClass,
			(('size', integerClass),),
			arrayCtor_int, False)
	])

#-----------------------------------------------------------------------------

# Mtodos de la clase Dictionary

def dictClass_extraMakePersistent(instance, manager):
	list = instance.getValue().values()
	for instance in list:
		instance.makePersistent(manager)

def dictClass_extraGetstate(instance, dict):
	value = instance.getValue().copy()
	newvalue = {}
	for key, instance in value.items():
		newvalue[key] = instance.getID()
	dict['value'] = newvalue

def dictClass_extraStore(instance, storage):
	dict = instance.getValue()
	for instance in dict.values():
		instance.store(storage)

def dictClass_extraRestore(instance, interpreter, storage):
	dict = instance.getValue()
	for key, id in dict.items():
		inst = storage.retrieveObject(id)
		inst.restore(interpreter, storage)
		dict[key] = inst

dictClass.extra_makePersistent = dictClass_extraMakePersistent
dictClass.extra_getstate = dictClass_extraGetstate
dictClass.extra_store = dictClass_extraStore
dictClass.extra_restore = dictClass_extraRestore
dictClass.updateFromParent()

def dictMethod_getItem(interpreter, method, instance, args):
	key = args[0].getInstance()
	if key.isNull():
		raise 'La clave en getItem() no puede ser nula'
	dict = instance.getValue()
	try:
		return dict[key.getValue()]
	except:
		return nullInstance # sera mejor lanzar una excepcin?

def dictMethod_setItem(interpreter, method, instance, args):
	key = args[0].getInstance()
	if key.isNull():
		raise 'La clave en setItem() no puede ser nula'
	value = args[1].getInstance()
	dict = instance.getValue()
	dict[key.getValue()] = value
	if instance.isPersistent():
		instance.manager.makePersistent(value)
	instance.setModified()
	return voidInstance

def dictMethod_delItem(interpreter, method, instance, args):
	key = args[0].getInstance()
	if key.isNull():
		raise 'La clave en delItem() no puede ser nula'
	dict = instance.getValue()
	try:
		del dict[key.getValue()]
	except:
		pass # sera mejor lanzar una excepcin?
	instance.setModified()
	return nullInstance

def dictMethod_hasItem(interpreter, method, instance, args):
	key = args[0].getInstance()
	if key.isNull():
		raise 'La clave en hasItem() no puede ser nula'
	dict = instance.getValue()
	return makeBool(key.getValue() in dict)

def dictMethod_clear(interpreter, method, instance, args):
	instance.getValue().clear()
	instance.setModified()
	return voidInstance

dictClass.addMethods([
		JMethod('getItem', objectClass, (('key', stringClass),),
			dictMethod_getItem),
		JMethod('setItem', voidClass,
			(('key', stringClass), ('item', objectClass)),
			dictMethod_setItem),
		JMethod('hasItem', boolClass, (('key', stringClass),),
			dictMethod_hasItem),
		JMethod('delItem', voidClass, (('key', stringClass),),
			dictMethod_delItem),
		JMethod('clear', voidClass, (),
			dictMethod_clear)
	])

# Constructores

def dictCtor_default(interpreter, ctor, instance, args):
	instance.setValue({})
	return voidInstance

dictClass.addConstructors([
		JConstructor(arrayClass, (),
			dictCtor_default, False)
	])

#-----------------------------------------------------------------------------

# Mtodos de la clase PersistenceManager

managerClass.updateFromParent()

def managerMethod_storeObject(interpreter, method, instance, args):
	arg = args[0].getInstance()
	if arg.isNull():
		raise 'El argumento de storeObject no puede ser nulo'
	manager = instance.getValue()
	manager.storeObject(arg)
	return makeSimple(stringClass, instance.getID())

def managerMethod_retrieveObject(interpreter, method, instance, args):
	arg = args[0].getInstance()
	if arg.isNull():
		raise 'El argumento de retrieveObject no puede ser nulo'
	id = arg.getValue()
	manager = instance.getValue()
	try:
		inst = manager.retrieveObject(id)
		return inst
	except KeyError, e:
		return nullInstance # Sera mejor lanzar una excepcin?

def managerMethod_setStorage(interpreter, method, instance, args):
	arg = args[0].getInstance()
	if arg.isNull():
		raise 'El argumento de setStorage no puede ser nulo'
	manager = instance.getValue()
	name = arg.getValue()
	manager.setStorage(name)
	return voidInstance

def managerMethod_setPolicy(interpreter, method, instance, args):
	arg = args[0].getInstance()
	if arg.isNull():
		raise 'El argumento de setPolicy no puede ser nulo'
	manager = instance.getValue()
	name = arg.getValue()
	manager.setPolicy(name)
	return voidInstance

managerClass.addMethods([
		JMethod('storeObject', stringClass, (('obj', objectClass),),
			managerMethod_storeObject),
		JMethod('retrieveObject', objectClass, (('id', stringClass),),
			managerMethod_retrieveObject),
		JMethod('setStorage', objectClass, (('name', stringClass),),
			managerMethod_setStorage),
		JMethod('setPolicy', objectClass, (('name', stringClass),),
			managerMethod_setPolicy)
	])

# Constructores

def managerCtor_default(interpreter, ctor, instance, args):
	instance.setValue(interpreter.getPersistenceManager())
	return voidInstance

managerClass.addConstructors([
		JConstructor(managerClass, (),
			managerCtor_default, False)
	])
